package dev.fitko.fitconnect.client.strategies;

import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.submission.*;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.*;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.client.model.AttachmentPayload;
import dev.fitko.fitconnect.client.model.DataPayload;
import dev.fitko.fitconnect.client.model.SubmissionPayload;
import dev.fitko.fitconnect.core.util.StopWatch;
import org.apache.tika.Tika;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URLConnection;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Collectors;

import static dev.fitko.fitconnect.client.util.SubmissionUtil.*;

public class SendNewSubmissionStrategy implements SubmitStrategy {

    private static final Logger LOGGER = LoggerFactory.getLogger(SendNewSubmissionStrategy.class);

    private final Sender sender;

    public SendNewSubmissionStrategy(final Sender sender) {
        this.sender = sender;
    }

    @Override
    public Optional<SentSubmission> send(final SubmissionPayload submissionPayload) {

        final UUID destinationId = submissionPayload.getDestinationId();
        final ServiceType serviceType = submissionPayload.getServiceType();

        try {

            final RSAKey encryptionKey = this.sender.getEncryptionKeyForDestination(destinationId);
            final Destination destination = this.sender.getDestination(destinationId);

            final List<AttachmentPayload> encryptedAttachments = encryptAndHashAttachments(encryptionKey, submissionPayload.getAttachments());
            final CreateSubmission newSubmission = buildSubmissionToAnnounce(destinationId, serviceType, encryptedAttachments);
            final SubmissionForPickup announcedSubmission = this.sender.createSubmission(newSubmission);

            final var startTimeAttachmentUpload = StopWatch.start();
            uploadAttachments(encryptedAttachments, announcedSubmission.getSubmissionId());
            LOGGER.info("Uploading attachments took {}", StopWatch.stopWithFormattedTime(startTimeAttachmentUpload));

            final DataPayload encryptedDataPayload = encryptDataPayload(submissionPayload.getData(), destination,  encryptionKey);
            final Metadata metadata = buildMetadataToSend(serviceType, encryptedAttachments, encryptedDataPayload);

            final ValidationResult validatedMetadata = this.sender.validateMetadata(metadata);
            if (validatedMetadata.hasError()) {
                LOGGER.error("Metadata does not match schema", validatedMetadata.getError());
                return Optional.empty();
            }

            final String encryptedData = encryptedDataPayload.getEncryptedData();
            final String encryptedMetadata = this.sender.encryptObject(encryptionKey, metadata);
            final SubmitSubmission submitSubmission = buildSubmitSubmission(announcedSubmission.getSubmissionId(), encryptedData, encryptedMetadata);

            final var startTimeSubmissionUpload = StopWatch.start();
            final Submission submission = this.sender.sendSubmission(submitSubmission);
            LOGGER.info("Uploading submission took {}", StopWatch.stopWithFormattedTime(startTimeSubmissionUpload));

            LOGGER.info("SUCCESSFULLY HANDED IN SUBMISSION !");
            return buildSentSubmission(encryptedAttachments, encryptedData, encryptedMetadata, submission);

        } catch (final EncryptionException e) {
            LOGGER.error("Encrypting submission failed", e);
        } catch (final RestApiException e) {
            LOGGER.error("Sending submission failed", e);
        } catch (final SchemaNotFoundException e) {
            LOGGER.error("Required schema to send valid submission not found", e);
        } catch (final SubmissionNotCreatedException e) {
            LOGGER.error("Failed to announce new submission", e);
        } catch (final KeyNotRetrievedException e) {
            LOGGER.error("Getting encryption key for destination {} failed", destinationId, e);
        } catch (final InvalidKeyException e) {
            LOGGER.error("Key for destination {} is invalid", destinationId, e);
        } catch (final AttachmentCreationException e) {
            LOGGER.error("Reading file failed. Attachment will not be created.", e);
        }
        return Optional.empty();
    }


    private DataPayload encryptDataPayload(final DataPayload dataPayload, final Destination destination, final RSAKey encryptionKey) {
        final String hashedData = this.sender.createHash(dataPayload.getRawData());
        final String encryptedData = this.sender.encryptBytes(encryptionKey, dataPayload.getRawData());
        final URI schemaUri = getSchemaUri(dataPayload, destination);
        return dataPayload.withSchemaUri(schemaUri)
                .withEncryptedData(encryptedData)
                .withHashedData(hashedData);
    }

    private URI getSchemaUri(final DataPayload dataPayload, final Destination destination) {
        final List<SubmissionSchema> submissionSchemas = getSubmissionSchemasFromDestination(destination);
        final String supportedSchemas = submissionSchemas.stream().map(s -> s.getMimeType().toString()).collect(Collectors.joining(" | "));
        final Optional<URI> schemaUriForMimeType = getSchemaUriForMimeType(submissionSchemas, dataPayload.getMimeType());
        if (schemaUriForMimeType.isEmpty()) {
            LOGGER.error("Destination supports the mime-type '{}'. Mime-type {} is not allowed, please check the destination", supportedSchemas, dataPayload.getMimeType());
            throw new SchemaNotFoundException("Schema for mime-type " + dataPayload.getMimeType() + " not found");
        }
        return schemaUriForMimeType.get();
    }

    private void uploadAttachments(final List<AttachmentPayload> attachmentPayloads, final UUID submissionId) {
        if (attachmentPayloads.isEmpty()) {
            LOGGER.info("No attachments to upload");
        } else {
            LOGGER.info("Uploading {} attachment(s)", attachmentPayloads.size());
        }
        for (final AttachmentPayload payload : attachmentPayloads) {
            this.sender.uploadAttachment(submissionId, payload.getAttachmentId(), payload.getEncryptedData());
        }
    }

    private List<AttachmentPayload> encryptAndHashAttachments(final RSAKey encryptionKey, final List<AttachmentPayload> attachments) {
        return attachments.stream().filter(Objects::nonNull).map(payload -> encryptAndHashAttachment(encryptionKey, payload)).collect(Collectors.toList());
    }

    private AttachmentPayload encryptAndHashAttachment(final RSAKey encryptionKey, final AttachmentPayload attachmentPayload) {
        final File file = attachmentPayload.getFile();
        try {
            final byte[] rawData = Files.readAllBytes(Paths.get(file.getPath()));
            final String encryptedAttachment = this.sender.encryptBytes(encryptionKey, rawData);
            final String hashedBytes = this.sender.createHash(rawData);
            return attachmentPayload.withEncryptedData(encryptedAttachment)
                    .withHashedData(hashedBytes)
                    .withMimeType(setMimeType(attachmentPayload));
        } catch (final Exception e) {
            throw new AttachmentCreationException("Attachment '" + file.getAbsolutePath() + "' could not be created ", e);
        }
    }

    private String setMimeType(final AttachmentPayload attachmentPayload) {
        final File file = attachmentPayload.getFile();
        String mimeType;
        try {
            final Tika mimeTypeGuesser = new Tika();
            mimeType = mimeTypeGuesser.detect(file);
        } catch (final IOException e) {
            mimeType = URLConnection.guessContentTypeFromName(file.getName());
        }
        LOGGER.info("Detected attachment mime-type {}", mimeType);
        return mimeType;
    }
}
